
// RangeSlider.java
// Ernie Yu, December 2010, 

/* The original version of RangeSlider can be found at
       http://ernienotes.wordpress.com/2010/12/27/creating-a-java-swing-range-slider/

   I added RangeSliderPanel, and simplified the RangeSlider() construction
   by making a few small changes to this class.
*/

/* An extension of JSlider to select a range of values using two thumb controls.
   The thumb controls are used to select the lower and upper value of a range
   with predetermined minimum and maximum values.
   
   Note that RangeSlider makes use of the default BoundedRangeModel, which 
   supports an inner range defined by a value and an extent.  The upper value
   returned by RangeSlider is simply the lower value plus the extent.
 */


package rslider;

import java.awt.*;
import javax.swing.JSlider;


public class RangeSlider extends JSlider
{
  /* Constructs a RangeSlider with default minimum and maximum values of 0
     and 100.  */
  public RangeSlider()
  {  initSlider();  }


  // Constructs a RangeSlider with the specified default minimum and maximum values
  public RangeSlider(int min, int max)
  { super(min, max);
    initSlider();
  }


  public RangeSlider(int min, int max, int lower, int upper)   // added by Andrew
  { super(min, max);
    initSlider();
    setUpperValue(upper);
    setValue(lower);
  }


  // Initializes the slider by setting default properties.
  private void initSlider() 
  { setOrientation(HORIZONTAL);
    setPreferredSize(new Dimension(240, getPreferredSize().height)); // added by Andrew
  }


  // Overrides the superclass method to install the UI delegate to draw two thumbs
  @Override
  public void updateUI()
  {
    setUI(new RangeSliderUI(this));
    // Update UI for slider labels.  This must be called after updating the
    // UI of the slider.  Refer to JSlider.updateUI().
    updateLabelUIs();
  }


  // Returns the lower value in the range
  @Override
  public int getValue()
  {  return super.getValue();  }


  // Sets the lower value in the range
  @Override
  public void setValue(int value)
  {
    int oldValue = getValue();
    if (oldValue == value)
      return;

    // Compute new value and extent to maintain upper value
    int oldExtent = getExtent();
    int newValue = Math.min(Math.max(getMinimum(), value), oldValue + oldExtent);
    int newExtent = oldExtent + oldValue - newValue;

    // Set new value and extent, and fire a single change event.
    getModel().setRangeProperties(newValue, newExtent, getMinimum(), 
                                    getMaximum(), getValueIsAdjusting());
  }  // end of setValue()


  // Returns the upper value in the range
  public int getUpperValue()
  {  return getValue() + getExtent();   }


  // Sets the upper value in the range
  public void setUpperValue(int value)
  {
    // Compute new extent
    int lowerValue = getValue();
    int newExtent = Math.min(Math.max(0, value - lowerValue),
        getMaximum() - lowerValue);
    // Set extent to set upper value
    setExtent(newExtent);
  }  // end of setUpperValue()


}  // end of RangeSlider class
